from .sound import ULSound2D
from uuid import uuid4
from uplogic.utils import debug, interpolate
from uplogic.events import schedule_callback


class ULMusicEffect():
    def __init__(
        self
    ):
        self._volume = 1.0
        self._pitch = 1.0
        self._fade_event = None

    @property
    def volume(self):
        return self._volume

    @volume.setter
    def volume(self, val):
        self._volume = val
    
    def fade_out(self, factor=.02):
        self.volume = interpolate(self.volume, 0, factor)
        if self._fade_event:
            self._fade_event.cancel()
            self._fade_event = None
        if self.volume > 0:
            self._fade_event = schedule_callback(self.fade_out, arg=factor)
    
    def fade_in(self, factor=.01):
        self.volume = interpolate(self.volume, 1, factor)
        if self._fade_event:
            self._fade_event.cancel()
            self._fade_event = None
        if self.volume < 1:
            self._fade_event = schedule_callback(self.fade_in, arg=factor)


class ULMusic(ULMusicEffect):
    def __init__(
        self,
        name: str = ''
    ):
        super().__init__()
        self.name = name if name else uuid4()
        self.tracks: list[ULMusicTrack] = []

    @property
    def position(self):
        if not self.tracks:
            return 0.0
        return self.tracks[0].sound.position

    @position.setter
    def position(self, val):
        for track in self.tracks:
            track.position = val

    @property
    def volume(self):
        return self._volume

    @volume.setter
    def volume(self, val):
        self._volume = val
        for track in self.tracks:
            track.sound.volume = track.volume * val
        if self._fade_event:
            self._fade_event.cancel()
            self._fade_event = None

    @property
    def pitch(self):
        return self._volume

    @pitch.setter
    def pitch(self, val):
        self._pitch = val
        for track in self.tracks:
            track.sound.pitch = track.pitch * val
        if self._fade_event:
            self._fade_event.cancel()
            self._fade_event = None

    def add_track(
        self,
        sound: str or ULSound2D,
        track_name: str = ''
    ):
        if not track_name:
            track_name = uuid4()
        track = ULMusicTrack(self, sound, track_name)
        self.tracks.append(track)
        return track

    def remove_track(
        self,
        track: int or str = 0,
    ):
        if isinstance(track, str):
            for t in self.tracks:
                if t.name == track:
                    t.remove()
                    return
            debug(f'Track "{track}" not found!')
        else:
            self.tracks[track].remove()

    def get_track(self, name):
        for track in self.tracks:
            if track.name == name:
                return track

    def pause(self):
        for track in self.tracks:
            track.sound.pause()

    def resume(self):
        for track in self.tracks:
            track.sound.resume()
    
    def stop(self):
        for track in self.tracks:
            track.sound.stop()
        self.position = 0.0


class ULMusicTrack(ULMusicEffect):
    def __init__(
        self,
        music: ULMusic,
        sound: str or ULSound2D,
        name: str
    ):
        super().__init__()
        self.music = music
        self.name = name
        sound = ULSound2D(sound) if isinstance(sound, str) else sound
        sound.position = music.position
        self.sound: ULSound2D = sound

    @property
    def position(self):
        return self.sound.position
    
    @position.setter
    def position(self, val):
        self.sound.position = val

    @property
    def volume(self):
        return self._volume
    
    @volume.setter
    def volume(self, val):
        self._volume = val
        self.sound.volume = val * self.music.volume
        if self._fade_event:
            self._fade_event.cancel()
            self._fade_event = None

    @property
    def pitch(self):
        return self._pitch
    
    @pitch.setter
    def pitch(self, val):
        self._pitch = 1.0
        self.sound.pitch = val * self.music.pitch

    def remove(self):
        self.sound.stop()
        self.music.tracks.remove(self)



from bge import logic
from bge.types import KX_GameObject as GameObject
import bpy
from bpy.types import Material


def create_curve(
    name: str,
    bevel_depth: float = 0.0,
    dimensions: int = 3,
    material: str or Material = None,
    collection: str = None
) -> GameObject:
    """Set the curve points of a `KX_GameObject` containing a `bpy.types.Curve` object.
    """
    bcurve = bpy.data.curves.new(name, 'CURVE')
    bcurve.bevel_depth = bevel_depth
    bcurve.dimensions = f'{dimensions}D'
    bobj = bpy.data.objects.new(name, bcurve)
    if material:
        if isinstance(material, str):
            bobj.data.materials.append(bpy.data.materials[material])
        elif isinstance(material, Material):
            bobj.data.materials.append(material)
    if collection:
        if isinstance(collection, str):
            collection = bpy.data.collections.get(collection, bpy.context.scene.collection)
    elif collection is None:
        collection = bpy.context.scene.collection
    collection.objects.link(bobj)
    game_obj = logic.getCurrentScene().convertBlenderObject(bobj)
    return game_obj


def set_curve_points(
    curve: GameObject,
    points: list
) -> None:
    """Set the curve points of a `KX_GameObject` containing a `bpy.types.Curve` object.
    """
    bcurve = curve.blenderObject.data
    for spline in bcurve.splines:
        bcurve.splines.remove(spline)
    spline = bcurve.splines.new('POLY')
    pos = curve.worldPosition

    spline.points.add(len(points)-1)
    for p, new_co in zip(spline.points, points):
        p.co = ([
            new_co[0] - pos.x,
            new_co[1] - pos.y,
            new_co[2] - pos.z
        ] + [1.0])


class ULCurve():
    """Wrapper class for creating and handling curves more easily.
    """

    def __init__(
        self,
        name: str,
        bevel_depth: float = 0.0,
        dimensions: int = 3,
        material: str or Material =None,
        collection: str = None
    ) -> None:
        self.object = create_curve(
            name,
            bevel_depth,
            dimensions,
            material,
            collection
        )

    @property
    def points(self):
        splines = self.object.blenderObject.data.splines
        return splines[0].points if len(splines) > 0 else []

    @points.setter
    def points(self, val):
        if val != self.points:
            set_curve_points(self.object, val)

    @property
    def bevel_depth(self):
        return self.object.blenderObject.data.bevel_depth

    @bevel_depth.setter
    def bevel_depth(self, val):
        self.object.blenderObject.data.bevel_depth = val